#-*- encoding: utf-8 -*-

import inspect
from time import strftime, localtime
import time
import sys
import traceback

__author__ = "Damian Kęska"
__license__ = "LGPLv3"
__maintainer__ = "Damian Kęska"
__copyright__ = "Copyleft by Panthera Desktop Team"

class pantheraLogging:
    """
        Simple logging and debugging class for Panthera Desktop Framework
        
    """

    logging = True
    silent = False
    session = ""
    panthera = ""
    lastLogTime = 0
    lastMsgTime = 0
    dateFormat = 'number'
    
    
    
    def initializeTimer(self):
        """
            Initialize timer for benchmarking custom code blocks
        """
    
        self.lastLogTime = time.time()
    
    
    
    def convertMessage(self, message, stackPosition):
        """
            Format a message
            
            message - message
            stackPosition - function name on stack
            
            Returns string
        """
    
        diff = "-"
    
        if self.lastLogTime > 0:
            diff = str(time.time()-self.lastLogTime)+ " real"
        elif self.lastMsgTime > 0:
            diff = str(time.time()-self.lastMsgTime)
            
        self.lastMsgTime = time.time()
        diff = diff[0:10]

        timeStr = str(time.time())

        if self.dateFormat != 'number':
            timeStr = time.strftime(self.dateFormat, time.localtime(int(float(timeStr))))

        return "["+timeStr+", "+str(diff)+"] "+stackPosition+": "+message
        
        
    
    def __init__ (self, panthera, quiet=False):
        """ 
            Constructor, takes panthera object as argument 
        """
    
        self.panthera = panthera
        
        if quiet:
            self.silent = True
        
    def outputException(self, *args, **kwargs):
        self.output(*args, **kwargs)
        traceback.print_exc(file=sys.stdout)

    def output(self, message, group='', savetoLogs=True, execHook=True, skipDate=False):
        """ 
            Outputs to log file and to console 
        """
        
        if self.logging == False:
            return False
            
        if not skipDate:
            message = self.convertMessage(message, inspect.stack()[1][3])
            
        if self.silent == False:
            print(message)
            
        self.session += message + "\n"

    def flushAndEnablePrinting(self, clearSession=False):
        """
        Flush buffers and re-enable log printint to screen
        :param clearSession:
        :return:
        """

        print(self.session[0:-1])

        if clearSession:
            self.session = ""

        self.silent = False



import imp
import os
import sys

try:
    from PySide import QtCore
except ImportError:
    from PyQt4 import QtCore

def bytes2human(n, format="%(value)i%(symbol)s"):
    """
    >>> bytes2human(10000)
    '9K'
    >>> bytes2human(100001221)
    '95M'
    """
    
    symbols = ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
    prefix = {}
    for i, s in enumerate(symbols[1:]):
        prefix[s] = 1 << (i+1)*10
    for symbol in reversed(symbols[1:]):
        if n >= prefix[symbol]:
            value = float(n) / prefix[symbol]
            return format % locals()
    return format % dict(symbol=symbols[0], value=n)

def human2bytes(s):
    """
    >>> human2bytes('1M')
    1048576
    >>> human2bytes('1G')
    1073741824
    """
    
    symbols = ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
    letter = s[-1:].strip().upper()
    num = s[:-1]
    assert num.isdigit() and letter in symbols
    num = float(num)
    prefix = {symbols[0]:1}
    for i, s in enumerate(symbols[1:]):
        prefix[s] = 1 << (i+1)*10
    return int(num * prefix[letter])

def include(path):
    """
    Include a Python file
    :param path: Path to .py file
    :return:
    """
    fileName, fileExtension = os.path.splitext(os.path.basename(path))
    return imp.load_source(os.path.basename(fileName).replace('.', ''), path)


def _decode_list(data):
    rv = []
    for item in data:
        if isinstance(item, unicode):
            item = item.encode('utf-8')
        elif isinstance(item, list):
            item = _decode_list(item)
        elif isinstance(item, dict):
            item = _decode_dict(item)
        rv.append(item)
    return rv


def _decode_dict(data):
    rv = {}
    for key, value in data.iteritems():
        if isinstance(key, unicode):
            key = key.encode('utf-8')
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        elif isinstance(value, list):
            value = _decode_list(value)
        elif isinstance(value, dict):
            value = _decode_dict(value)
        rv[key] = value
    return rv

class pantheraWorker(QtCore.QObject):
    """
        Worker for pantheraWorkThread
    """

    finished = QtCore.pyqtSignal()
    dataReady = QtCore.pyqtSignal(list, dict)
    job = None
    jobArgs = None
    thread = None

    def setJob(self, job, args='', thread=''):
        self.job = job
        self.jobArgs = args
        self.thread = thread

    """
        Run callable function inside of thread
    """

    def run(self):
        if self.job:
            if self.jobArgs:
                self.job(self.jobArgs, self.thread)
            else:
                self.job(thread=self.thread)
        else:
            print("Warning: pantheraWorker job was not set")

        self.finished.emit()

class pantheraWorkThread(QtCore.QThread):
    def __init__(self, parent=None):
        QtCore.QThread.__init__(self, parent)

     # this class is solely needed for these two methods, there
     # appears to be a bug in PyQt 4.6 that requires you to
     # explicitly call run and start from the subclass in order
     # to get the thread to actually start an event loop

    def start(self):
        QtCore.QThread.start(self)

    def run(self):
        QtCore.QThread.run(self)

def createThread(callable, args='', autostart=True):
    """ Create a Worker and Thread and connect them """

    appThread = pantheraWorkThread()
    appWorker = pantheraWorker()
    appWorker.setJob(callable, args, appThread)
    appWorker.moveToThread(appThread)
    appWorker.finished.connect(appThread.terminate)
    appThread.started.connect(appWorker.run)

    if autostart:
        appThread.start()

    return appThread, appWorker

def daemonize (stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
    '''This forks the current process into a daemon.
    The stdin, stdout, and stderr arguments are file names that
    will be opened and be used to replace the standard file descriptors
    in sys.stdin, sys.stdout, and sys.stderr.
    These arguments are optional and default to /dev/null.
    Note that stderr is opened unbuffered, so
    if it shares a file with stdout then interleaved output
    may not appear in the order that you expect.
    '''

    # Do first fork.
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0)   # Exit first parent.
    except OSError as e:
        sys.stderr.write ("fork #1 failed: (%d) %s\n" % (e.errno, e.strerror) )
        sys.exit(1)

    # Decouple from parent environment.
    os.chdir("/")
    os.umask(0)
    os.setsid()

    # Do second fork.
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0)   # Exit second parent.
    except OSError as e:
        sys.stderr.write ("fork #2 failed: (%d) %s\n" % (e.errno, e.strerror) )
        sys.exit(1)

    # Redirect standard file descriptors.
    si = open(stdin, 'r')
    so = open(stdout, 'a+')
    se = open(stderr, 'a+', 0)
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())


class pantheraClass:
    """ Panthera class """

    panthera = ""

    def __init__(self, panthera):
        """ Initialize object """
    
        self.panthera = panthera
        self.app = panthera
        self.main()
        
    def main(self):
        return None


from pantheradesktop import pantheraClass
import pantheradesktop.tools as tools
import sys
import os
from code import InteractiveConsole

class Console(InteractiveConsole):
    def __init__(*args): InteractiveConsole.__init__(*args)

class pantheraInteractiveConsole(pantheraClass):
    thread = None
    worker = None

    ## put here your console shortcuts
    recognizedChars = {
        'q': 'pa_exit'
    }

    def main(self):
        """
        Main function that registers base shortcuts and starts a thread
        :return:
        """

        self.recognizedChars['q'] = self.app.pa_exit
        self.recognizedChars['l'] = self.clearConsole
        self.recognizedChars['p'] = self.pythonConsole
        self.thread, self.worker = tools.createThread(self.interactiveConsole)


    def clearConsole(self):
        """
        Clear console screen
        :return:
        """

        os.system('clear')


    def pythonConsole(self):
        """
        Starts an interactive Python console
        :return:
        """

        app = panthera = self.app

        vars = globals().copy()
        vars.update(locals())
        console = Console(vars)
        console.interact()


    def interactiveConsole(self, thread=''):
        """
        Interactive console loop thread
        :param thread:
        :return:
        """

        while True:
            char = sys.stdin.read(1)

            if char and char in self.recognizedChars:
                self.recognizedChars[char]()



#-*- encoding: utf-8 -*-
import os
import sys
import atexit
import pantheradesktop.config
import pantheradesktop.hooking
import pantheradesktop.logging
import pantheradesktop.tools as tools
import pantheradesktop.instance
import pantheradesktop.interactive
import traceback

try:
    import pantheradesktop.interactive
except Exception:
    pass

try:
    import pantheradesktop.argsparsing
except Exception:
    pass

try:
    import pantheradesktop.qtgui
except Exception:
    pass

import pantheradesktop.db

__author__ = "Damian Kęska"
__license__ = "LGPLv3"
__maintainer__ = "Damian Kęska"
__copyright__ = "Copyleft by Panthera Desktop Team"

class Singleton(object):
    _instance = None
    
    def __new__(class_, *args, **kwargs):
        if not isinstance(class_._instance, class_):
            class_._instance = object.__new__(class_, *args, **kwargs)
            
        return class_._instance

class pantheraDesktopApplication(Singleton):
    """
       Panthera Desktop Framework
       
       Main class used to be extended by desktop application, provides base functionality (configuration, database, templating)
       
    """

    config = "" # cofiguration object
    interactive = ""
    template = "" # gui template object
    argsParser = ""
    instances = ""
    app = ""
    __appMain = ""
    _plugins = {}
    pluginsLoaded = False
    pluginsAvailable = []

    # application name
    appName = "pantheradesktop-exampleapp"
    version = "0.1"
    
    # directory where to store data eg. ~/.example (will be automaticaly generated in initialize function)
    filesDir = ""
    
    # core classes
    coreClasses = {
        'hooking': pantheradesktop.hooking.pantheraHooking, 
        'logging': pantheradesktop.logging.pantheraLogging, 
        'argsparsing': pantheradesktop.argsparsing.pantheraArgsParsing, 
        'config': pantheradesktop.config.pantheraConfig,
        'gui': pantheradesktop.qtgui.pantheraQTGui, # set to None to disable
        'db': pantheradesktop.db.pantheraDB,
        'instances': '__initialize__',
        'interactive': '__initialize__'
    }
    
    def multipleIsFile(self, dirs, fileName):
        for path in dirs:
            if os.path.isfile(path+"/"+fileName):
                return path+"/"+fileName
                
        return False
    
    @classmethod
    def getInstance(c):
        return c._instance
    
    def initialize(self, quiet=False):
        """
            Create required directories, initialize basic objects
            
        """

        if self.coreClasses['instances'] == '__initialize__' and hasattr(pantheradesktop, 'instance'):
            self.coreClasses['instances'] = pantheradesktop.instance.pantheraJSONInstance

        if self.coreClasses['interactive'] == '__initialize__' and hasattr(pantheradesktop, 'interactive'):
            self.coreClasses['interactive'] = pantheradesktop.interactive.pantheraInteractiveConsole
        
        atexit.register(self.pa_exit)
        self.filesDir = os.path.expanduser("~/."+self.appName)

        # plugins paths
        self.pluginsSearchDirectories = [
            self.filesDir+'/plugins',
            '/usr/share/'+self.appName+'/plugins',
            '/var/lib/'+self.appName+'/plugins'
        ]
        
        # create user's data directory if missing
        if not os.path.isdir(self.filesDir):
            try:
                os.mkdir(self.filesDir)
            except Exception as e:
                print("Cannot create "+self.filesDir+" directory, please check permissions (details: "+e.strerror+")")
                sys.exit(5)

        self.hooking = self.coreClasses['hooking']()
        self.logging = self.coreClasses['logging'](self, quiet)
        
        # plugins support: action before configuration load
        self.hooking.execute('app.beforeConfigLoad')
               
        if not os.path.isfile(self.filesDir+"/config.json"):
            try:
                w = open(self.filesDir+"/config.json", "w")
                w.write("{}")
                w.close()
            except Exception as e:
                print("Cannot create "+self.filesDir+"/config.json, please check permissions (details: "+e.strerror+")")
                sys.exit(5)       
        
        # initialize configuration
        self.config = self.coreClasses['config'](self)
        self.config.configPath = self.filesDir+"/config.json"

        # initialize database
        if self.coreClasses['db']:
            self.db = self.coreClasses['db'](self)


    def togglePlugin(self, pluginName, value=None):
        """
        Toggle plugin on/off
        :param pluginName: Plugin name
        :return: bool
        """

        if not self.pluginsLoaded == True:
            self.loadPlugins()

        plugins = self.config.getKey('plugins', list())

        if (value == False or (value == None and pluginName in plugins)) and pluginName in plugins:
            plugins.remove(pluginName)
        elif value == True or (value == None and pluginName not in plugins):
            plugins.append(pluginName)

        self.config.setKey('plugins', plugins)
        self.config.save()

        return True

    def loadPlugins(self, force=False):
        """
        Load application plugins from directories specified in self.pluginsSearchDirectories
        Only plugins enabled in configuration will be loaded
        :return:
        """

        if self.pluginsLoaded == True and not force:
            return False

        # create a default value for "plugins" key
        self.config.getKey('plugins', list())
        self.config.save()

        self.logging.output('Looking for plugins in: '+str(self.pluginsSearchDirectories), 'pantheraDesktop.plugins')

        for path in self.pluginsSearchDirectories:

            # check only directories that exists in filesystem
            if not os.path.isdir(path):
                continue

            files = os.listdir(path)

            for file in files:
                fileName = os.path.basename(file)
                fileName, fileExtension = os.path.splitext(file)

                if fileExtension.lower() != ".py" or fileName in self._plugins:
                    continue

                # append to list of avaliable plugins
                self.pluginsAvailable.append(fileName)

                if self.config.getKey('plugins') and not fileName in self.config.getKey('plugins'):
                    self.logging.output('Disabling plugin '+fileName, 'pantheraDesktop')
                    continue

                try:
                    self.loadPlugin(fileName, path+"/"+file)

                except Exception as e:
                    self.logging.output('Cannot initialize plugin '+path+'/'+file+', details: '+str(e), 'pantheraDesktop')
                    self.logging.output(traceback.format_exc(), 'pantheraDesktop')

        self.pluginsLoaded = True


    def loadPlugin(self, pluginName, path=None):
        """
        Initialize plugin
        :param pluginName: Plugin name eg. "backup" or "notify"
        :param path: Path to plugins directory /usr/share/copysync/plugins
        :return: object
        """

        if not path:
            for directory in self.pluginsSearchDirectories:
                if pluginName+'.py' in os.listdir(directory):
                    path = directory
                    break

        if not path:
            raise IOError('No such file or directory', 2)

        plugin = tools.include(path)
        self.logging.output('Initializing plugin '+pluginName, 'pantheraDesktop')
        self._plugins[pluginName] = eval("plugin."+pluginName+"Plugin(self)")

        ## call initializePlugin() method if available
        if hasattr(self._plugins[pluginName], 'initializePlugin'):
            self._plugins[pluginName].initializePlugin()

        return self._plugins[pluginName]


    def unloadPlugin(self, pluginName):
        """
        Unload plugin
        :param pluginName: Plugin name eg. "backup" or "notify"
        :return: bool
        """

        if not pluginName in self._plugins:
            return False

        ## execute pre-unload function if exists to clean up some stuff
        if hasattr(self._plugins[pluginName], 'unloadPlugin'):
            self.logging.output('Calling "'+pluginName+'" plugin to deinitialize itself')
            self._plugins[pluginName].unloadPlugin()

        ## try to remove all hooks defined by plugin we want to remove
        self.hooking.removeAllByClass(self._plugins[pluginName].__class__.__name__)

        del self._plugins[pluginName]
        return True


    def main(self, func=None):
        """ Main function """
        
        # validate "instances" module
        if 'instances' in self.coreClasses and callable(self.coreClasses['instances']) and hasattr(self.coreClasses['instances'], 'register'):
            self.logging.output('Registering self instance using '+str(self.coreClasses['instances'])+' instance handler')
            self.instances = self.coreClasses['instances'](self)
            self.instances.register()
            self.instances.cleanup()
            self.hooking.addOption('app.pa_exit', self.instances.unregister)

        if 'interactive' in self.coreClasses and callable(self.coreClasses['interactive']):
            self.logging.output('Initializing interactive mode', 'pantheraDesktop')
            self.interactive = self.coreClasses['interactive'](self)

        # initialize plugins
        self.loadPlugins()

        # initialize args parser
        self.argsParser = self.coreClasses['argsparsing'](self)
        self.argsParser.parse()
        self.hooking.execute('app.argsparsing.init')

        # add default value for configAutocommit
        self.config.getKey('configAutocommit', True)

        self.logging.output('Initializing application mainloop', 'pantheraDesktopApplication')
        
        # graphical user interface (if avaliable)
        if "gui" in self.coreClasses:
            if self.coreClasses['gui']:
                self.gui = self.coreClasses['gui'](self)
        
        # plugins support: mainloop
        self.hooking.execute('app.mainloop')
        
        if hasattr(func, '__call__') or "classobj" in str(type(func)):
            self.__appMain = func(self)


    def pa_exit(self):
        """ On application exit """

        self.logging.output('Got interrupt, finishing jobs and exiting...', 'pantheraDesktop')
        self.hooking.execute('app.pa_exit')
        sys.exit(0)



#-*- encoding: utf-8 -*-
from pantheradesktop import pantheraClass
import os
import pwd
import json
import copy
import sys
import subprocess

# workaround for bug: OSError: [Errno 2] No such file or directory when trying to use os.getlogin()
os.getlogin = lambda: pwd.getpwuid(os.getuid())[0]

class pantheraJSONInstance(pantheraClass):
    pid = None

    def register(self, params=''):
        """
        Register instance of application to instances.json file so other instances or applications could know about runtime of this application
        :param params: Dictionary of parameters
        :hook panthera.instance.parameters: Dictionary of parameters
        :function self.registerParams: Dictionary of parameters
        :return: bool
        """

        instancesFile = self.app.filesDir+'/instances.json'

        ## if file does not exists then create an empty file
        if not os.path.isfile(instancesFile):
            w = open(instancesFile, 'w')
            w.write("{\n}")
            w.close()

        if type(params).__name__ is not "dict":
            params = dict()

        ## get self process id
        self.pid = str(os.getpid())

        params['process.user.login'] = os.getlogin()
        params['process.user.id'] = os.getuid()
        params['sys.argv'] = sys.argv

        ## add support for extensions
        if hasattr(self, 'registerParams'):
            params = self.registerParams(params)

        params = self.app.hooking.execute('panthera.instance.parameters', params)

        instancesList = json.loads(open(instancesFile).read())
        instancesList[self.pid] = params

        self.__writeToJSONFile(instancesList)
        return True

    def cleanup(self):
        """
        Clean up instances list that crashed and does not exists anymore
        :return: bool
        """

        processTable = subprocess.check_output('ps aux', shell=True)
        instancesList = json.loads(open(self.app.filesDir+'/instances.json').read())

        if not instancesList:
            return True

        newInstancesList = copy.copy(instancesList)

        for instanceID,data in instancesList.iteritems():
            if not " "+str(instanceID)+" " in processTable:
                del newInstancesList[str(int(instanceID))]

        self.__writeToJSONFile(newInstancesList)
        return True


    def unregister(self, hook=''):
        """
        Unregister self from instances.json
        :return: bool
        """


        if not self.pid:
            return False

        instancesList = json.loads(open(self.app.filesDir+'/instances.json').read())

        if str(self.pid) in instancesList:
            del instancesList[str(self.pid)]

        self.__writeToJSONFile(instancesList)
        return True


    def __writeToJSONFile(self, data):
        w = open(self.app.filesDir+'/instances.json', 'w')
        w.write(json.dumps(data, sort_keys=True, indent=4, separators=(',', ': ')))
        w.close()


import json
import sys
import pantheradesktop.tools as tools

class pantheraConfig:
    """
        Panthera Desktop Configuration
        
        Simple json based key-value storage
        
    """

    panthera = ""
    memory = {}
    configPath = ""
    configurationChanged = False
    strictTypeChecking = True

    def __init__(self, panthera):
        self.panthera = panthera
        self.panthera.hooking.addOption('app.pa_exit', self.save, 1)

    def loadConfig(self):
        """
            Load configuration from JSON file to memory
            
        """

        self.panthera.logging.output("Loading configuration from path " + self.configPath, "pantheraConfig")

        try:
            t = open(self.configPath, "rb")
            self.memory = json.loads(t.read(), object_hook=tools._decode_dict)
        except Exception as e:
            print("Cannot parse configuration file \"" + self.configPath + "\"")
            sys.exit(5)  # errno.EIO = 5

        t.close()


    def getKey(self, key, defaultValue = None, strictTypeChecking = False):
        """
            Get configuration key
        
            key - name
            defaultValue - default value to set if key does not exists yet
        
        """

        if not self.memory:
            self.loadConfig()

        ## if strict value type checking is on we will stick with one data type
        if defaultValue is not None and (self.strictTypeChecking or strictTypeChecking) and key in self.memory:

            if str(type(defaultValue).__name__) != str(type(self.memory[key]).__name__):
                if self.panthera.logging:
                    self.panthera.logging.output('Invalid data type for "'+key+'" key ('+type(defaultValue).__name__+', '+type(self.memory[key]).__name__+')', 'pantheraConfig')

                # set default value in case we have an unexpected type set in configuration file
                self.setKey(key, defaultValue)


        if key in self.memory:
            return self.memory[key]

        ## if key does not exists in key-value database yet, create it with default value
        if defaultValue is not None:
            self.setKey(key, defaultValue)

        return defaultValue


    def setKey(self, key, value):
        """
            Set configuration key

            key - name
            value - value

        """

        if type(key) is not str:
            return False

        if type(value) is type(sys) or type(value) == object:
            return False

        self.configurationChanged = True
        self.memory[key] = value

        if self.getKey('configAutocommit'):
            self.save()


    def exists(self, key):
        """
        Check if key exists
        :param key:
        :return:
        """

        return key in self.memory


    def save(self):
        """
            Save configuration right back to json file

        """

        if self.configurationChanged:
            w = open(self.configPath, "wb")
            w.write(json.dumps(self.memory, sort_keys=True, indent=4, separators=(',', ': ')))
            w.close()


#-*- encoding: utf-8 -*-

import pantheradesktop.tools as tools
import argparse
import sys
import os

__author__ = "Damian Kęska"
__license__ = "LGPLv3"
__maintainer__ = "Damian Kęska"
__copyright__ = "Copyleft by Panthera Desktop Team"

class pantheraArgsParsing:
    """
        Arguments parsing class for Panthera Desktop Framework
        It's just a simple wrapper to argparse module that supports callbacks
        
    """

    description = 'Example app'
    argparse = ""
    panthera = None
    app = None
    
    # here will be args and opts stored
    args = None
    opts = None

    # plugins management
    _enablePluginsManagement = True
    _enableConfigManagement = True
    _enableDebugArgs = True
    _enableDaemonizeArgs = True

    knownArgs = {}

    def __init__ (self, panthera):
        """
            Initialize argument parser with Panthera Framework object
            
            panthera - Panthera object
        """

        self.panthera = panthera
        self.app = panthera

        self.appendStaticArguments()
        self.argparse = argparse.ArgumentParser(description=self.description)
        self.createArgument('--version', self.version, 'Display help', action='store_true', required=False)

        if self._enablePluginsManagement:
            self.createArgument('--plugins', self.pluginsList, 'Display plugins list', action='store_true', required=False)
            self.createArgument('--plugins-enable', self.pluginsEnable, 'Enable plugin', required=False)
            self.createArgument('--plugins-disable', self.pluginsDisable, 'Enable plugin', required=False)

        if self._enableConfigManagement:
            self.createArgument('--config', self.configKeys, 'List configuration keys', action='store_true', required=False)
            self.createArgument('--config-set', self.configSetKey, 'Set configuration key value', required=False)
            self.createArgument('--config-get', self.configGetKey, 'Get configuration value', required=False)

        if self._enableDebugArgs:
            self.createArgument('--debug', self.setDebuggingMode, '', 'Enable debugging mode', required=False, action='store_false')

        if self._enableDaemonizeArgs:
            self.createArgument('--daemonize', self.setDaemonMode, '', 'Enable debugging mode', required=False, action='store_false')

        # @hook app.argsparsing.__init__.after self
        self.app.hooking.execute('app.argsparsing.__init__.after', self)


    def appendStaticArguments(self):
        """
        Append static arguments from configuration key "panthera.argsparsing.defaultArgs"

        :return: None
        """

        if self.app.config.getKey('panthera.argsparsing.defaultArgs', ''):
            sys.argv = sys.argv + self.app.config.getKey('panthera.argsparsing.defaultArgs').split(' ')

    def setDaemonMode(self, opt = ''):
        """
        Run application in daemon mode
        :param opt:
        :return:
        """

        if not os.path.isdir(self.panthera.filesDir + '/logs'):
            os.mkdir(self.panthera.filesDir + '/logs')

        tools.daemonize(stdout = self.panthera.filesDir + '/logs/stdout', stderr = self.panthera.filesDir + '/logs/stderr')


    def setDebuggingMode(self, opt = ''):
        """
        Enable debugging mode
        :param opt:
        :return:
        """

        self.app.logging.silent = False
        self.app.logging.flushAndEnablePrinting()


    def configKeys(self, opt=''):
        """
        List all configuration keys
        :param opt:
        :return:
        """

        self.app.config.loadConfig()
        print(self.app.config.memory)
        sys.exit(0)

    def configSetKey(self, key=''):
        """
        Set a configuration key
        :param key: Configuration key name or xpath
        :return:
        """

        self.app.config.loadConfig()

        if not len(self.opts):
            sys.exit(1)

        self.app.config.setKey(key, self.opts[0])
        self.app.config.save()

        sys.exit(0)

    def configGetKey(self, key=''):
        """
        Get configuration key
        :param key: Key name or xpath
        :return:
        """

        self.app.config.loadConfig()

        if self.app.config.exists(key):
            print(self.app.config.getKey(str(key)))
            sys.exit(0)
        else:
            sys.exit(1)

    def pluginsEnable(self, pluginName='', enable=True):
        """
        Enable a plugin
        :param pluginName:
        :param enable:
        :return:
        """

        self.app.loadPlugins()

        if not pluginName in self.app.pluginsAvailable:
            print("Cannot toggle plugin: "+str(pluginName)+", file not found")
            sys.exit(1)

        self.app.togglePlugin(pluginName, enable)
        sys.exit(0)

    def pluginsDisable(self, pluginName):
        """
        Disable a plugin
        :param pluginName:
        :return:
        """

        return self.pluginsEnable(pluginName, False)

    def pluginsList(self, opt=''):
        """
        List all plugins (enabled and disabled)
        :param opt:
        :return:
        """

        self.app.loadPlugins()

        print("Application is looking for plugins in those directories: ")

        for directory in self.app.pluginsSearchDirectories:
            print(directory)

        print("\n")

        if self.app.pluginsAvailable:
            print("Available plugins: ")

            for plugin in self.app.pluginsAvailable:
                if plugin in self.app._plugins:
                    print("[x] "+plugin)
                else:
                    print("[ ] "+plugin)

            print("\nx - enabled")
        else:
            print("No plugins available\n")

        sys.exit(0)
        
    def version(self, value=''):
        """
            Example argument handler, shows application version
            
        """

        self.app.hooking.execute('app.argsparsing.version', self)
        print(self.panthera.appName + " " +self.panthera.version)
        sys.exit(0)
        
        
        
    def createArgument(self, arg, callback, default="", help="", required=False, dataType=None, action='store', choices=None, type=None, skipArgParse=False, short=None):
        """
            Add argument to list of known arguments
            
            arg - Argument name eg. --test
            callback - Function to call back if this argument was used
            default - Default value
            help - Help text to display
            required - Is this a required field?
            skipArgParse - Skip using argparse in this function and use it by your own
            
            Returns none
        """
        
        self.knownArgs[arg] = callback
        argsString = 'arg, default=default, help=help, required=required, action=action'

        if not skipArgParse:
            if not action == 'store_true' and not action == 'store_const' and not action == 'store_false':
                argsString += ', choices=choices, type=type'

            if short is not None:
                argsString = 'short, '+argsString

            eval('self.argparse.add_argument('+argsString+')')

        
    def parse(self):
        """
            Run arguments parsing
            
        """
        
        if "addArgs" in dir(self):
            self.addArgs()

            # @hook app.argsparsing.parse.addArgs self
            self.app.hooking.execute('app.argsparsing.parse.addArgs', self)
        
        self.args = self.argparse.parse_known_args()
        self.opts = self.args[1]

        # @hook app.argsparsing.parse.before self
        self.app.hooking.execute('app.argsparsing.parse.before', self)

        for arg in self.knownArgs:
            if arg in sys.argv:
                self.knownArgs[arg](self.args[0].__dict__[arg.replace('--', '').replace('-', '_')])

        # @hook app.argsparsing.parse.after self
        self.app.hooking.execute('app.argsparsing.parse.after', self)


#-*- encoding: utf-8 -*-
import os
import sys
import json

try:
    from PySide import QtGui, QtCore
except ImportError:
    from PyQt4 import QtGui, QtCore

__author__ = "Damian Kęska"
__license__ = "LGPLv3"
__maintainer__ = "Damian Kęska"
__copyright__ = "Copyleft by Panthera Desktop Team"

class pantheraQTGui:
    panthera = ""
    templatesDir = []
    templates = {}
    app = ""
    startupTemplate = "main" # template to load at startup (leave empty to don't load any template at startup)

    def __init__(self, panthera):
        """
            Initialize class with Panthera object
        """

        self.panthera = panthera
        
        if os.path.isdir("/usr/share/"+self.panthera.appName+"/qtgui"):
            self.templatesDir.append("/usr/share/"+self.panthera.appName+"/qtgui")
        
        if os.path.isdir(os.getcwd()+"/usr"):
            self.templatesDir.append("./usr/share/"+self.panthera.appName+"/qtgui")
            
        self.templatesDir.append(self.panthera.filesDir+"/qtgui")
            
        for path in self.templatesDir:
            sys.path.append(path)
            
        self.app = QtGui.QApplication(sys.argv)
        
        # connect mainloop hook
        self.panthera.hooking.addOption('app.mainloop', self.mainLoop)
        
        
            
    def loadTemplate(self, templateName):
        """
            Compiles UI, save to cache, import cache and driver
            
        """

        # check for source XML file exported from Qt Designer
        uiFile = self.panthera.multipleIsFile(self.templatesDir, "/src/"+templateName+".ui")
        
        if not uiFile:
            raise Exception('Cannot find template source file "' +templateName + '.ui" in templatesDir = '+json.dumps(self.templatesDir)+'/src')
            return False

        # search for compiled Python PyQt4 code     
        cacheFile = self.panthera.multipleIsFile(self.templatesDir, "/cache/"+templateName+".py")
               
        if not cacheFile or os.path.getmtime(uiFile) > os.path.getmtime(cacheFile): # if there is no cache file or it's outdated
            for path in self.templatesDir:
                try:
                    # try to compile the file using pyuic4 and place in writable directory
                    os.system("pyuic4 -x "+uiFile+" > "+path+"/cache/"+templateName+".py")
                except Exception as e:
                    pass
                    
                if os.path.isfile(path+"/cache/"+templateName+".py"):
                    self.panthera.logging.output("Compiled template to '"+path+"/cache/"+templateName+".py'", 'pantheraQTGui')
                    break
    
        try:
            exec("import cache."+templateName+" as templateModule")
        except Exception as e:
            raise Exception('Cannot load template '+templateName+'.py from any path provided in templatesDir = '+json.dumps(self.templatesDir)+'/cache')
            return False # just in case
            
        try:
            exec("import drivers."+templateName+" as templateDriver")
        except Exception as e:
            raise Exception('Cannot load template driver '+templateName+'.py from any path provided in templateDir = '+json.dumps(self.templatesDir)+'/drivers (Exception: '+str(e)+')')
            
        self.templates[templateName] = {'module': templateModule, 'driver': templateDriver.templateMain(templateModule, self.panthera)}
        return self.templates[templateName]['driver']
         
    def mainLoop(self, data=''):
        """
            Main loop hook
            
        """
        
        if self.startupTemplate:
            template = self.loadTemplate(self.startupTemplate)
            template.show()

        try:
            sys.exit(self.app.exec_())
            sys.exit(0)
        except KeyboardInterrupt:
            sys.exit(0)        


class pantheraTemplateDriver:
    viewModule = ""
    view = ""
    panthera = ""

    def __init__(self, view, panthera):
        self.viewModule = view
        self.panthera = panthera
        
        for attribute in dir(view):
            if "Ui_" in attribute:
                self.panthera.logging.output('Found GUI class "'+attribute+'"', 'pantheraTemplateDriver')
                exec("self.view = view."+attribute+"()")
                self.initializeView()
                
    def initializeView(self):
        """
            Initialize view
        """

        pass
                
    def show(self):
        self.panthera.logging.output('Showing QMainWindow', 'pantheraTemplateDriver')
        self.window = QtGui.QMainWindow()
        self.view.setupUi(self.window)
        self.window.show()



#-*- encoding: utf-8 -*-

import inspect
import time
import traceback

__author__ = "Damian Kęska"
__license__ = "LGPLv3"
__maintainer__ = "Damian Kęska"
__copyright__ = "Copyleft by Panthera Desktop Team"

class pantheraHooking:
    """
        Hooking (plugins support) module for Panthera Desktop Framework

    """

    hooksList = {}
    
    def addOption(self, hookName, func, priority=99):
        """
            Connect method to a hooking slot
            
            hookName - hooking slot name
            func - function address
            priority - 0-99 priority number (lowest = higher priority)
            
            Returns bool
        """
        
        if priority > 99 or priority < 0:
            raise Exception('Priority should be in range of 0 to 99')
            return False

        # create array if there is no any hooked function yet
        if not hookName in self.hooksList:
            self.hooksList[hookName] = {}

        argsData = False

        try:
            args = inspect.getargspec(func).args

            if (len(args) > 0 and args[0] != 'self') or (len(args) > 1 and args[0] == 'self'):
                argsData = True

        except Exception:
            pass
            
        self.hooksList[hookName][str(priority)+'_'+str(func.__str__)] = {
            'callback': func,
            'argsData': argsData,
            'priority': priority,
            'created': time.time(),
            'executed': 0
        }
        
        return True


    def removeAllByClass(self, objectClassName):
        """
        Remove all hooks that reffers to selected object type
        :param objectClassName:
        :return: int Number of affected hooks
        """

        i = 0

        for hookName, hooks in self.hooksList.iterkeys():
            for id, hook in hooks.iterkeys():
                if hasattr(hook['callback'], 'im_class') and hook['callback'].im_class.__name__ == objectClassName:
                    self.removeOption(hook['callback'], hookName)
                    i = i + 1

        return i

        
    def execute(self, hookName, data=''):
        """
            Execute all functions on selected slot
            
            hookName - hooking slot name
            data - data to pass
            
            Returns data
        """
    
        if not hookName in self.hooksList:
            return data

        for func in sorted(self.hooksList[hookName]):
            try:
                # get data

                if self.hooksList[hookName][func]['argsData']:
                    data = self.hooksList[hookName][func]['callback'](data)
                else:
                    data = self.hooksList[hookName][func]['callback']()

                # increase counter
                self.hooksList[hookName][func]['executed'] = self.hooksList[hookName][func]['executed'] + 1
            except Exception as e:
                print('Hook: '+hookName)
                print('Data: '+str(data))
                #raise RuntimeError('Cannot execute hooked function: "'+str(self.hooksList[hookName][func]['callback'])+'" in "'+hookName+'" slot, error details: '+str(e), 1)
                traceback.print_exc()
            
        return data


    def hookExists(self, hookName):
        """
            Check if hooking slot exists
            
            hookName - hooking slot name
            
            Returns bool
        """

        return hookName in self.hooksList


    def removeOption(self, func, hookName=None):
        """
            Remove an option from hooking slot
            
            func - function
            hookName - hooking slot name
            
            Returns bool
        """
    
        return isConnected(func, hookName, True)


    def isConnected(self, func, hookName=None, removeHook=False):
    
        """
            Check if selected function is connected to selected or any hook slot
            
            func - function
            hookName - optional hooking slot
            
            Returns bool
        """
        
        for name in self.hooksList:
        
            if hookName is not None:
                if not self.hookExists(hookName):
                    return False
            
                name = self.hooksList[hookName]
        
            for hook in name:
                if hook['callback'] == func:
                    if removeHook:
                        del self.hooksList[hookName]
                
                    return True
    
            if hookName is not None:
                return False



#-*- encoding: utf-8 -*-

__author__ = "Damian Kęska"
__license__ = "LGPLv3"
__maintainer__ = "Damian Kęska"
__copyright__ = "Copyleft by Panthera Desktop Team"

import sys
import codecs
import traceback

try:
    import MySQLdb
    import MySQLdb.cursors
except ImportError:
    pass
    
try:
    import sqlite3
except ImportError:
    pass
    
try:
    from peewee import *
    import peewee
except ImportError:
    pass

try:
    from pymongo import MongoClient
except ImportError:
    pass

class pantheraDB:
    """
        Database support
    """
    
    panthera = None
    db = None
    cursor = None
    dbType = None
    escapeFunc = None

    def __init__(self, panthera):
        """ Initialize connection """
    
        self.panthera = panthera
        
        self.panthera.logging.output("Initializing pantheraDB", "pantheraDB")
        
        # Create some default configuration keys
        self.panthera.config.getKey('databaseHost', 'localhost')
        self.panthera.config.getKey('databasePassword', '')
        self.panthera.config.getKey('databaseUser', 'root')
        self.panthera.config.getKey('databaseDB', 'my-database')
        self.panthera.config.getKey('databaseFile', self.panthera.filesDir+"/db.sqlite3")
        
        
        if not "MySQLdb" in globals() and not "sqlite3" in globals() and not "peewee" in globals():
            self.panthera.logging.output("No MySQL or SQLite3 driver found", "pantheraDB")
            sys.exit(1)
            
        try:
            dbType = self.panthera.config.getKey('databaseType', 'orm').lower()
            
            ###
             #
             # Peewee ORM database
             #
            ###
            
            if dbType == 'orm':
                if not "peewee" in globals():
                    self.panthera.logging.output("Peewee module not found, but orm database type selected", "pantheraDB")
                    sys.exit(1)
                
                if self.panthera.config.getKey("databaseSocketType", "sqlite3") == "sqlite3":
                    self.db = db = SqliteDatabase(self.panthera.config.getKey('databaseFile', self.panthera.filesDir+"/db.sqlite3"))
                    
                    class BaseModel(Model):
                        class Meta:
                            database = db
                    
                    self.BaseModel = BaseModel
                    self.dbType = "peewee"
                
            ###
             #
             # SQLite3 native support
             #
            ###
            
            elif dbType == 'sqlite3':
                self.db = sqlite3.connect(self.panthera.config.getKey('databaseFile'), check_same_thread=False)
                self.db.row_factory = dict_factory
                self.cursor = self.db.cursor()
                self.dbType = "sqlite3"
                self.escapeFunc = quoteIdentifier
                
            ###
             #
             # MySQL (MySQLdb driver)
             #
            ###
                
            elif dbType == 'mysql':
                
                self.db = MySQLdb.connect(
                    host=self.panthera.config.getKey('databaseHost'),
                    user=self.panthera.config.getKey('databaseUser'),
                    passwd=self.panthera.config.getKey('databasePassword'),
                    db=self.panthera.config.getKey('databaseDB'),
                    cursorclass=MySQLdb.cursors.DictCursor
                )
                
                # escape function
                self.escapeFunc = self.db.escape_string
                
                self.cursor = self.db.cursor()
                self.dbType = "mysql"
                
            ###
             #
             # MongoDB
             #
            ###
            
            elif dbType == 'mongodb':
                self.db = MongoClient(str(self.panthera.config.getKey('databaseHost')), int(self.panthera.config.getKey('databasePort', 27017)))
                
                if self.panthera.config.getKey('databaseUser') and self.panthera.config.getKey('databasePassword'):
                    self.db.the_database.authenticate(self.panthera.config.getKey('databaseUser'), self.panthera.config.getKey('databasePassword'))
                
                self.dbType = "mongodb"
            else:
                self.panthera.logging.output("Unknown database driver \'"+str(self.panthera.config.getKey('databaseType'))+"\'", "pantheraDB")
                sys.exit(1)
                
            self.panthera.logging.output("Connection estabilished using "+self.dbType+" socket", "pantheraDB")
        except Exception as e:
            self.panthera.logging.outputException("Cannot connect to database: "+str(e), "pantheraDB")
            sys.exit(1)


    def query(self, query, values=dict(), commit=True):
        """ Execute a raw query """
        
        ###
         # Append database prefix and insert values into a query
        ###

        # {$db_prefix} insertion support
        query = query.replace('{$db_prefix}', str(self.panthera.config.getKey('databasePrefix', 'pa_')))
        originalQuery = query

        # inserting escaped values into query
        query, values = self.applyValues(query, values)

        if query.strip() != originalQuery.strip():
            self.panthera.logging.output("Original: " + query, "pantheraDB")

        self.panthera.logging.output("SQL: "+ query, "pantheraDB")

        ###
         # Make q query and return a resultset
        ###
        
        if self.dbType == "peewee":
            return self.db.execute_sql(query, values)
            
        elif self.dbType == "sqlite3":
            obj = self.cursor.execute(query, values)
            
            if commit:
                self.db.commit()
                
            return pantheraDBSQLite3ResultSet(obj, self.cursor)
        
        elif self.dbType == 'mysql':
            obj = self.cursor.execute(query, values)
            
            if commit:
                self.db.commit()
                
            return pantheraDBMySQLResultSet(self.cursor, self, obj)
        
        elif self.dbType == 'mongodb':
            raise Exception('query() is not supported by MongoDB database type')
        else:
            self.panthera.logging.output("Cannot connect to databse via unknown socket", "pantheraDB")
            sys.exit(1)



    def applyValues(self, query, values):
        """ Append values from dict to query string """

        newValues = []

        for value in values:
            pos = query.find(':' + value)

            while pos != -1:
                query = query[:pos] + '?' + query[pos + len(':' + value):]
                newValues.append(values[value])

                pos = query.find(':' + value)

        return query, newValues



class pantheraDBSQLite3ResultSet:
    """ Result set for SQLite3 """

    db = None
    cursor = None
    lastrowid = None
    indexColumn = None

    def __init__(self, cursor, db, result=''):
        self.db = db
        self.cursor = cursor
        self.lastrowid = 0
        
        if cursor:
            self.lastrowid = cursor.lastrowid
        
    def indexColumn(self, columnName):
        """ Column to index by """
    
        self.indexColumn = columnName
        return self
        
    def rowCount(self):
        """ Count affected/selected rows """
        
        if not self.cursor:
            return 0
    
        rowCount = self.cursor.rowcount
        
        if rowCount < 0:
            rowCount = 0
        
        return rowCount
        
    def fetchAll(self):
        """ Fetch all rows """
        
        if not self.cursor:
            return list()
    
        f = self.cursor.fetchall()
        
        if self.indexColumn is not None and len(f) > 0 and f[0].has_key(self.indexColumn):
            newArray = dict()
        
            for row in f:
                newArray[row[self.indexColumn]] = row
                
            f = newArray
    
        return f
        
    def fetch(self):
        """ Fetch one row """
        
        if not self.cursor:
            return False
    
        return self.cursor.fetchone()

    
class pantheraDBMySQLResultSet(pantheraDBSQLite3ResultSet):
    """ Result set for SQLite3 """

    db = None
    cursor = None
    lastrowid = None
    indexColumn = None
    
def quoteIdentifier(s, errors="strict"):
    encodable = s.encode("utf-8", errors).decode("utf-8")

    nul_index = encodable.find("\x00")

    if nul_index >= 0:
        error = UnicodeEncodeError("NUL-terminated utf-8", encodable,
                                   nul_index, nul_index + 1, "NUL not allowed")
        error_handler = codecs.lookup_error(errors)
        replacement, _ = error_handler(error)
        encodable = encodable.replace("\x00", replacement)

    return "\"" + encodable.replace("\"", "\"\"") + "\""
            
def dict_factory(cursor, row):
    """ Dictionary factory for SQLite3 """

    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d



#!/usr/bin/env python
#-*- encoding: utf-8 -*-

from distutils.core import setup

setup(name='pantheradesktop',
      description = "Panthera Desktop Framework",
      long_description = "Tiny desktop framework for easy application development in Python using PyQT/PySide",
      author = "Damian Kęska",
      author_email = "webnull.www@gmail.com",
      version="0.1.0.3",
      license = "LGPL",
      url = 'ttps://github.com/Panthera-Framework/Panthera-Desktop/',
      download_url = 'https://github.com/Panthera-Framework/Panthera-Desktop/archive/master.tar.gz',
      package_dir={'': 'src'},      
      packages=['pantheradesktop'],
      keywords=['panthera', 'desktop', 'framework', 'shell', 'apps', 'cli apps'],
      data_files = []
     )
